/*
 **************************************************************************
 *		SWI-Prolog ActiveX Extensions.
 *
 *	This software is provided in source and binary form free of charge to 
 *  users of SWI-Prolog.  No warranty of any kind is expressed or implied;
 *  in particular, this is entirely unsupported software.
 *
 *  Documentation for this product is available from the SWI-Prolog 
 *  web site (http://www.swi.psy.uva.nl/projects/SWI-Prolog/).
 *
 *	Version 1.0.1.
 *
 *	October 1, 1999
 *
 **************************************************************************
 */

//
//	plcvt.cpp
//
#include "proqcvt.h"

//  Keyword enum for common atoms
enum EKEYWD
{
	ENONE,
	ETRUE,
	EFALSE,
	EOPTIONAL,
	EMAX
};


EKEYWD eKeywordCheck ( term_t t )
{
	char * pszAtom;
	if ( ! PL_get_atom_chars( t, & pszAtom ) )
		return ENONE;

	if ( ::strcmp( "true", pszAtom ) == 0 )
		return ETRUE;
	else
	if ( ::strcmp( "false", pszAtom ) == 0 )
		return EFALSE;
	else
	if ( ::strcmp( "optional", pszAtom ) == 0 )
		return EOPTIONAL;

	return ENONE;
}

//
//	Class Id and guid conversion functions
//
bool bStringFromGuid ( const GUID & guid, ZSTR & zsGuid )
{
	LPOLESTR pszGUID = NULL;
	HRESULT hr = ::StringFromCLSID( guid, &pszGUID );
	if ( FAILED(hr) )
		return false;
	zsGuid = pszGUID;
	::CoTaskMemFree(pszGUID);
	return true;
}

bool bGuidFromString ( TSZC tszcGuid, GUID & guid )
{
	ZSTRW zswGuid( tszcGuid );
	HRESULT hr = ::CLSIDFromString( const_cast<WCHAR*>(zswGuid.Tszc()), & guid );
	return SUCCEEDED(hr);
}

bool bGetClsId ( TSZC tszcProgId, CLSID & clsid )
{
	ZSTRW zswProgId( tszcProgId );
	LONG lResult = CLSIDFromProgID( const_cast<WCHAR*>(zswProgId.Tszc()), & clsid );
	return lResult == 0;
}

bool bGetClsIdString ( TSZC tszcProgId, ZSTR & zsClsId )
{
	CLSID clsid;
	if ( ! bGetClsId( tszcProgId, clsid ) )
		return false;
	return bStringFromGuid( clsid, zsClsId );
}

bool bConvertVariantToInt ( _variant_t & vnt, int & iResult )
{	
	iResult = 0;
	bool bResult = true;
	switch ( vnt.vt )
	{		
		case VT_UI2:
		{
			iResult = vnt.uiVal;
			break;
		}

	    case VT_I2:
		{
			iResult = vnt.iVal;
			break;
		}

		case VT_UI4:
		case VT_UINT:
		{
			iResult = vnt.ulVal;
			break;
		}

		case VT_INT:
		case VT_I4:
		{
			iResult = vnt.lVal;
			break;
		}

		case VT_BOOL:
		{
			iResult = vnt.iVal == VARIANT_TRUE;
			break;
		}

		case VT_EMPTY:
			iResult = 0;
			break;

		default:
			bResult = false;
			break;
			
	}
	return bResult;
}

bool bPrepareList ( 
	term_t tList, 
	VARTYPE & vt, 
	int & citems )
{
	vt = VT_EMPTY;
	citems = 0;
	term_t tHead = PL_new_term_ref();
	term_t tTail = PL_copy_term_ref(tList);

	while ( PL_get_list( tTail, tHead, tTail ) )
	{
		VARTYPE vtItem = VT_NULL;
		//  Get the type of this term
		int iarg = PL_term_type( tHead );

		if ( PL_is_string( tHead ) )
			vtItem = VT_BSTR;
		else
		if ( PL_is_atom( tHead ) )
			vtItem = VT_BSTR;
		else
		if ( PL_is_integer( tHead ) )
			vtItem = VT_I4;
		else
		if ( PL_is_float( tHead ) )
			vtItem = VT_R8;
		else
		//BUGBUG! Force anything else to be a string
		if ( ! PL_is_variable( tHead ) )
			vtItem = VT_BSTR;	

		//  Can we handle this type? 
		//		(don't allow unbound variables in lists)
		if ( vtItem == VT_NULL )
			return false;

		//  If list is still of unassigned type, 
		if ( vt == VT_EMPTY )
			vt = vtItem;		// use this type
		else	// else, verify that it's of the same type
		if ( vt != vtItem )
			return false;

		citems++;
	}
	return true;
}

//
//	Convert a flat list of Prolog terms to a SAFEARRAY nestled in a VARIANT.
//	We do not allow nested lists, nor do we allow unbound terms in a list.
//
bool bConvertListToVariant ( term_t tList, _variant_t & vnt )
{
	int citems = 0;
	VARTYPE vt = VT_EMPTY;

	//  Check that all the elements of the list are of the same type
	if ( ! bPrepareList( tList, vt, citems ) )
		return false;

	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].lLbound = 0;
	rgsabound[0].cElements = citems;

	vnt.vt = VT_ARRAY | vt;
	HRESULT hr = S_OK;
	SAFEARRAY * psfa = NULL;

	double dbl;
	int i;
	ZSTRW zsw;
	void * pv;
	do
	{
		SAFEARRAY * psfa = ::SafeArrayCreate( vt, 1, rgsabound );
		if ( ! psfa )
		{
			hr = E_FAIL;
			break;
		}
		vnt.parray = psfa;
		hr = ::SafeArrayAllocData( psfa );
		if ( FAILED(hr) )
			break;
		
		term_t tHead = PL_new_term_ref();
		term_t tTail = PL_copy_term_ref(tList);

		for ( long iIndex = 0; PL_get_list( tTail, tHead, tTail ); iIndex++ )
		{			
			switch ( vt )
			{
				case VT_BSTR:
				{
					char * psz;
					if ( ! PL_get_chars( tHead, & psz, CVT_ALL ) )
					{
						hr = E_FAIL;
						break;
					}
					zsw = psz;
					_bstr_t bstr = zsw;
					pv = bstr.copy();
					break;
				}
				case VT_I4:
					PL_get_integer( tHead, & i );
					pv = & i;
					break;

				case VT_R8:
					PL_get_float( tHead, & dbl );
					pv = & dbl;
					break;

				default:
					assert(false);
					hr = E_FAIL;
					break;
			}
			if ( FAILED(hr) )
				break;
			hr = ::SafeArrayPutElement( psfa, & iIndex, pv );
		}
	} 
	while (false);

	if ( FAILED(hr) )
	{
		if ( psfa )
			::SafeArrayDestroy( psfa );
		vnt.parray = NULL;
		vnt.Clear();
	}
	return SUCCEEDED(hr);
}

//
//	Convert the given Prolog term to the desired VARIANT type.  
//     	Coerce special forms like "true", "false" and "optional" to 
//     	their correct types.
//
bool bConvertTermToVariant ( int vt, term_t tArg, _variant_t & vnt )
{
	IP_IDispatch ipId;

	//  Check to see if the current term's atom is a "special" atom
	EKEYWD ekwd = eKeywordCheck( tArg );
	switch ( ekwd )
	{
		case ETRUE:
		case EFALSE:
		{
			vnt.vt = VT_BOOL;
			vnt.boolVal = ekwd == ETRUE 
							? VARIANT_TRUE 
							: VARIANT_FALSE;
			return true;
			break;
		}
		case EOPTIONAL:
		{
			vnt.vt = VT_ERROR;
			vnt.scode = DISP_E_PARAMNOTFOUND;
			return true;
			break;
		}
		//  Not a keyword
		default:
			break;
	}

	//  If the argument is non-specific, convert by term type
	if ( vt == VT_VARIANT )
	{
		//  Get the type of this term
		int iarg = PL_term_type( tArg );
		switch ( iarg )
		{
			//  Integers
			case PL_INTEGER:
				vt = VT_I4;
				break;

			//  Reals
			case PL_FLOAT:
				vt = VT_R8;
				break;

			//  See if this is an interface reference or a list
			case PL_TERM:
			//  Anything vaguely resembling a string
			case PL_ATOM:   
			case PL_STRING:
				vt = VT_BSTR;
				break;
			
			default:
				return false;
		}
	}

	switch ( vt )
	{
		//  Two-byte integers
		case VT_UI2:
	    case VT_I2:
		{
			int i;
			if ( ! PL_get_integer( tArg, & i ) )
				return false;
			short s = i;
			vnt = s;
			break;
		}

		case VT_INT:
		case VT_UI4:
		case VT_UINT:
		case VT_I4:
		{
			int i;
			if ( ! PL_get_integer( tArg, & i ) )
				return false;
			vnt = (long) i;
			break;
		}

		case VT_R4:
		{
			double d;
			if ( ! PL_get_float( tArg, & d ) )
				return false;
			vnt = (float)d;
			break;
		}

		case VT_R8:
		{
			double d;
			if ( ! PL_get_float( tArg, & d ) )
				return false;
			vnt = d;
			break;
		}

		case VT_BSTR:
		{
			char * pch = NULL;
			if ( ! PL_get_chars( tArg, & pch, CVT_ALL | BUF_DISCARDABLE ) )
				return false;

			_bstr_t bstr = pch;
			vnt = bstr;
			break;
		}

		case VT_BOOL:
		{
			char * pch = NULL;
			if ( ! PL_get_chars( tArg, & pch, CVT_ALL | BUF_DISCARDABLE ) )
				return false;
			bool b = ::_strcmpi("true",pch) == 0;
			vnt.Clear();
			new (& vnt) _variant_t( b );
			break;
		}

		default:
			return false;
	}
	return true;
}


//
//	Convert a list of Prolog terms to a SAFEARRAY nestled in a VARIANT.
//	We do not allow unbound terms in a list.  All elements of a list must
//	be of the same type unless the VARTYPE is VT_VARIANT.
//
bool bConvertListToVariantSafearray ( 
	term_t tList,						//  The list head
	_variant_t & vnt,					//	The resulting variant containing the SAFEARRAY
	VARTYPE vt /* = VT_EMPTY */ )		//	The desired VARTYPE; VT_EMPTY implies "guess"
{
	int citems = 0;

	//  Check that all the elements of the list are of the same type
	if ( ! bPrepareList( tList, vt, citems ) )
		return false;

	SAFEARRAYBOUND rgsabound[1];
	rgsabound[0].lLbound = 0;
	rgsabound[0].cElements = citems;

	vnt.vt = VT_ARRAY | vt;
	HRESULT hr = S_OK;
	SAFEARRAY * psfa = NULL;

	double dbl;
	int i;
	ZSTRW zsw;
	void * pv;
	do
	{
		SAFEARRAY * psfa = ::SafeArrayCreate( vt, 1, rgsabound );
		if ( ! psfa )
		{
			hr = E_FAIL;
			break;
		}
		vnt.parray = psfa;
		hr = ::SafeArrayAllocData( psfa );
		if ( FAILED(hr) )
			break;
		
		term_t tHead = PL_new_term_ref();
		term_t tTail = PL_copy_term_ref(tList);

		for ( long iIndex = 0; 
			  SUCCEEDED(hr) && PL_get_list( tTail, tHead, tTail ); 
			  iIndex++ )
		{			
			switch ( vt )
			{
				case VT_BSTR:
				{
					char * psz;
					if ( ! PL_get_chars( tHead, & psz, CVT_ALL ) )
					{
						hr = E_FAIL;
						break;
					}
					zsw = psz;
					_bstr_t bstr = zsw;
					pv = bstr.copy();
					break;
				}

				case VT_I1:
				case VT_UI1:
				case VT_I2:
				case VT_UI2:
				case VT_I4:
				case VT_UI4:
				case VT_INT:
				case VT_UINT:
				{
					PL_get_integer( tHead, & i );
					pv = & i;
					break;
				}
				case VT_R8:
				{
					PL_get_float( tHead, & dbl );
					pv = & dbl;
					break;
				}

				case VT_VARIANT:
				{
					//  This should never change
					assert( sizeof(VARIANT) == sizeof(_variant_t) );
					//  Construct a new _variant_t in the SAFEARRAY data
				    _variant_t * pvar = (_variant_t *) pv;
					new (pvar) _variant_t;
					//  Convert the next term to a variant
					if ( ! bConvertTermToVariant( VT_VARIANT, tHead, *pvar ) )
						hr = E_FAIL;
					break;
				}

				default:
				{
					assert(false);
					hr = E_FAIL;
					break;
				}
			}
			if ( SUCCEEDED(hr) )
				hr = ::SafeArrayPutElement( psfa, & iIndex, pv );
		}
	} 
	while (false);

	if ( FAILED(hr) )
	{
		if ( psfa )
			::SafeArrayDestroy( psfa );
		vnt.parray = NULL;
		vnt.Clear();
	}
	return SUCCEEDED(hr);
}
